#include "StdAfx.h"
#include "updownclient.h"
#include "emule.h"
#include "uploadqueue.h"
#include "Clientlist.h"
#include "SearchList.h"
#include <zlib/zlib.h>
#include "otherfunctions.h"
#include "emuleDlg.h"
#include "TransferWnd.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//<<< eWombat [SUI]
void CUpDownClient::GetSUIStateStr(CString &string)
{
	if (!credits)
	{
		string.Empty();
		return;
	}
	switch(credits->GetCurrentIdentState(GetIP()))
	{
	case IS_IDNEEDED:
		string=GetResString(IDS_SUI_IDNEEDED);
		return;
	case IS_IDENTIFIED:
		string=GetResString(IDS_SUI_IDENTIFIED);
		return;
	case IS_IDFAILED:
		string=GetResString(IDS_SUI_IDFAILED);
		return;
	case IS_IDBADGUY:
		string=GetResString(IDS_SUI_IDBADGUY);
		return;
	}
	string.Empty();
	return;
}

int CUpDownClient::GetSUISort()
{
	if (!credits)
		return 5;
	switch(credits->GetCurrentIdentState(GetIP()))
		{
		case IS_IDENTIFIED:
			return 4;
		case IS_IDNEEDED:
			return 3;
		case IS_IDBADGUY:
			return 2;
		case IS_IDFAILED:
			return 1;
		}
	return 5;
}

int CUpDownClient::GetBanSort(void) const
{
	return IsBanned()+(10*GetSnafuReason())+(2*IsNotSUI())+(4*IsSUIFailed());
}

CString CUpDownClient::GetBanStr(void) const
{
	if (IsSnafu())
		return GetSnafuReasonStr();
	else if (IsNotSUI())
		return _T("[S.U.I.]");
	else
		return (IsBanned() ? GetResString(IDS_YES):GetResString(IDS_NO));
}
//<<< eWombat [SUI]

// <<< eWombat [SNAFU]
void CUpDownClient::UnSnafu(bool bCheckTrust) 
{
	if (bCheckTrust && m_nSnafuReason != SNAFU_ACT && m_nSnafuReason != SNAFU_HASHCHG)
		return;
	ResetTrust();
	m_nSnafuReason=SNAFU_NONE;
	m_nSnafuTimer=0;
	theApp.emuledlg->transferwnd->queuelistctrl.RefreshClient(this);
	theApp.emuledlg->transferwnd->clientlistctrl.RefreshClient(this);
}

void CUpDownClient::DoSnafu(eSnafuReason reason ,bool bCheckTrust,bool bLog)
{
	if (!thePrefs.GetAntiSnafu() || m_bSnafuExclude)
		return;
	if (IsSnafu())
		{
		if (m_nTrust > 0)
			m_nTrust--;
		if (reason==SNAFU_ACT || reason == SNAFU_HASHCHG || reason == SNAFU_HASH || reason==SNAFU_MYHASH)
			m_nSnafuReason=reason;
		return;
		}
	if (bCheckTrust && m_nTrust > 0)
		{
		m_nTrust--;
		return;
		}

	//If it is on Upload-queue remove it
	if(GetUploadState() == US_UPLOADING || GetUploadState() == US_WAITCALLBACK || GetUploadState() == US_CONNECTING )
		theApp.uploadqueue->RemoveFromUploadQueue(this,USR_SNAFU);

	m_nSnafuReason   = reason;
	m_nSnafuTimer	 = ::GetTickCount();
	theApp.clientlist->AddSnafuClient(this);
	theApp.emuledlg->transferwnd->queuelistctrl.RefreshClient(this);
	theApp.emuledlg->transferwnd->clientlistctrl.RefreshClient(this);
	if (bLog)
		DoSnafuInfo();
}

void CUpDownClient::DoSnafuInfo()
{
	CString text;
	CString buffer;
	buffer = this->GetClientSoftVer();	
	text.Format("S.N.A.F.U. %s [%s:%u]",GetSnafuReasonStr(),ipstr(GetConnectIP()),GetUserPort());
	if (m_pszUsername!=NULL)
		text.AppendFormat(" User: %s",m_pszUsername);
	else if (HasValidHash())
		text.AppendFormat(" Hash: %s",md4str(GetUserHash()));
	if (!buffer.IsEmpty())
		text.AppendFormat(" Mod: %s",buffer);
	text+=_T(" banned");
	if (thePrefs.GetVerbose()) AddPhoenixLogLine(false, text);
}

CString	CUpDownClient::GetSnafuReasonStr() const
{
	if (!thePrefs.GetAntiSnafu())
		return "";

	switch(m_nSnafuReason)
	{
	case SNAFU_WOMBAT:
		return "Bad Wombat";
	case SNAFU_FRIENDSHAREMOD:
		return "Friendshare Mod";
	case SNAFU_NAME:
	case SNAFU_ACTION:
	case SNAFU_OPCODE:
		return "suspect behaviour";
	case SNAFU_NICK:
		return "suspect nickname";
	case SNAFU_MODSTR:
		return "suspect client version";
	case SNAFU_CLIENT:
		return "eDonkey client with eMule property";
	case SNAFU_HASHCHG:
		return "hash changer";
	case SNAFU_MYHASH:
		return "hash bastard";
	case SNAFU_ACT:
		return "hash thief (act)";
	case SNAFU_HASH:
		return "hash thief";
	case SNAFU_CLONE:
		return "clone (InfoPacket thief)";
	default:
		return "shit happens";
	}
}
void CUpDownClient::SnafuCheckMod(LPCTSTR pMod)
{
	if (!pMod)
		return;

	if (stristrex(pMod,"mison")!=NULL ||stristrex(pMod,"evort")!=NULL ||stristrex(pMod,"booster")!=NULL ||
		stristrex(pMod,"hardmule")!=NULL || stristrex(pMod,"eifen")!=NULL ||
		stristrex(pMod,"elfen")!=NULL ||stristrex(pMod,"imp delta")!=NULL ||stristrex(pMod,"father")!=NULL)
		m_bGoofy=true;

	if (!thePrefs.GetAntiSnafu())
		return;
	if (IsSnafu())
		return;
	//CString test=m_strMod;
	//test.MakeLower();
	if (m_bGoofy ||
		// [TPT] - Updated to morph		
		stristrex(pMod,"evortex")!=NULL||
		stristrex(pMod,"$motty")!=NULL||
		stristrex(pMod,"thunder")!=NULL||
		stristrex(pMod,"buzzfuzz")!=NULL||
		stristrex(pMod,"speed-unit")!=NULL||
		stristrex(pMod,"killians")!=NULL||
		stristrex(pMod,"element")!=NULL||
		stristrex(pMod,"")!=NULL||
		// [TPT] - Updated to morph		
		stristrex(pMod,"bomb")!=NULL||
		stristrex(pMod,"fick")!=NULL||
		stristrex(pMod,"dmx")!=NULL||
		stristrex(pMod,"fincan")!=NULL||
		stristrex(pMod,"dodgethis")!=NULL|| //Updated
		stristrex(pMod,"freeza")!=NULL||
		stristrex(pMod,"d-unit")!=NULL||
		stristrex(pMod,"nos")!=NULL||
		stristrex(pMod,"imperator")!=NULL||
		stristrex(pMod,"speedload")!=NULL||
		stristrex(pMod,"gt mod")!=NULL||
		stristrex(pMod,"egomule")!=NULL||
		stristrex(pMod,"darkmule")!=NULL||
		stristrex(pMod,"legolas")!=NULL||
		stristrex(pMod,"dm-")!=NULL|| //hotfix
		stristrex(pMod,"|x|")!=NULL||
		stristrex(pMod,"esladevil")!=NULL||
		stristrex(pMod,"dragon")!=NULL)
	{	
		if (thePrefs.GetVerbose()) AddPhoenixLogLine(false, "Suspect Mod detected: %s",pMod);
		DoSnafu(SNAFU_MODSTR,false,false);
	}
}

void CUpDownClient::SnafuCheckNick(LPCTSTR pNick)
{
	if (!thePrefs.GetAntiSnafu())
		return;
	if (pNick==NULL)
		return;
	if (IsSnafu())
		return;
	if(	
		(stristrex(pNick,"emule") && stristrex(pNick,"booster")) ||
		// [TPT] - Updated to morph		
		stristrex(pNick,"dodgethis")	|| 
		stristrex(pNick,"chief")		|| 
		stristrex(pNick,"$motty")		|| 
		stristrex(pNick,"emule-speed")  || 
		stristrex(pNick,"gate-emule")	|| 
		stristrex(pNick,"buzzfuzz")		|| 
		stristrex(pNick,"speed-unit")	|| 
		stristrex(pNick,"killians")		|| 
		stristrex(pNick,"pubsman")		|| 
		stristrex(pNick,"emule-element")|| 		
		// [TPT] - Updated to morph		
		stristrex(pNick,"pharao")		|| 
		stristrex(pNick,"reverse") 		||
		stristrex(pNick,"rammstein")	||
		stristrex(pNick,"leecha")		||
		stristrex(pNick,"energyfaker")	||
		stristrex(pNick,"edevil")		||
		stristrex(pNick,"darkmule")		||
		stristrex(pNick,"evortex")		||
		stristrex(pNick,"|evorte|x|")	||
		stristrex(pNick,"$gam3r$")		||
		stristrex(pNick,"g@m3r")		||
		stristrex(pNick,"$warez$")		||
		stristrex(pNick,"celinesexy"))
	{	
		if (thePrefs.GetVerbose()) AddPhoenixLogLine(false, "Suspect nickname: %s",pNick);
		DoSnafu(SNAFU_NICK,false,false);
	}
}
void CUpDownClient::SnafuCheckVersion()
{
	if (IsSnafu())
		return;
	if ((GetVersion()>589) && ((GetSourceExchangeVersion()>0) || (GetExtendedRequestsVersion()>0)) && (GetClientSoft()==SO_EDONKEY)) //Fake Donkeys
	{	
		DoSnafu(SNAFU_CLIENT,false,false);
	}
	if ( (GetClientSoft() == SO_EDONKEYHYBRID) &&  ( (GetSourceExchangeVersion()>0) || (GetExtendedRequestsVersion()>0))) //Fake Hybrids
	{
		DoSnafu(SNAFU_CLIENT,false,false);
	}
	if (m_bySupportSecIdent!=0 && (GetClientSoft()==SO_EDONKEYHYBRID || GetClientSoft()==SO_EDONKEY))
	{
		DoSnafu(SNAFU_CLIENT,false,false);
	}
}

ULONG CUpDownClient::SnafuNameCRC(char* tocalc)
{   
	return theApp.wombatlist->GetHashKey(tocalc);
}

//>>> eWombat [ACT]

// >>> eWombat [SNAFU]
#ifdef _DEBUGWOMBAT
void CUpDownClient::GetDebugStr(CString &string)
{
	string="...";//m_strCommunity;
}
#endif

void  CUpDownClient::UnBan()
{
	theApp.clientlist->AddTrackClient(this); // [TPT] - Patch	
	SetUploadState(US_NONE);

	ClearWaitStartTime();

	// [TPT] - Code Improvement
	for (POSITION pos = m_RequestedFiles_list.GetHeadPosition();pos != 0;)
	{
		Requested_File_Struct* cur_struct = m_RequestedFiles_list.GetNext(pos);
		cur_struct->badrequests = 0;
		cur_struct->lastasked = 0;	
	}
	m_nBanTimer = 0;
	m_bBan		= false;
	theApp.emuledlg->transferwnd->queuelistctrl.RefreshClient(this);
	theApp.emuledlg->transferwnd->clientlistctrl.RefreshClient(this);
}

bool CUpDownClient::IsBanned() const
{	
	return( m_bBan && (m_nDownloadState != DS_DOWNLOADING));
}

void CUpDownClient::Ban()
{
	theApp.clientlist->AddTrackClient(this); // [TPT] - Patch
	if ((!IsBanned()) && (thePrefs.GetVerbose()))
			AddDebugLogLine(false,GetResString(IDS_CLIENTBLOCKED),GetUserName());
	m_nBanTimer = ::GetTickCount();
	m_bBan		= true;
	theApp.clientlist->AddSnafuClient(this);	
	SetUploadState(US_BANNED); // [TPT] - Patch
	theApp.emuledlg->transferwnd->queuelistctrl.RefreshClient(this);
	theApp.emuledlg->transferwnd->clientlistctrl.RefreshClient(this);
}
